local Addon = _G[...]
local pieces = {}

local onEnter = function(self)
	self.sets.alpha = self.sets.alpha or 100
	self:SetAlpha(self.sets.alpha/100 or 1)
end

local onLeave = function(self)
	self.sets.fadeAlpha = self.sets.fadeAlpha or 100
	self:SetAlpha(self.sets.fadeAlpha/100 or 1)
end

local basics = {
	frameStrata = 1,
	frameLevel = 2,
	alpha = 100,
	fadeAlpha = 100,
}

--Basics that all widgets share
local function Hook(frame)


	if not frame.hooked then
		frame.hooked = true
	else
		return
	end
	local def = frame.defaults
	if def then
		for i, b in pairs(basics) do
			def[i] = def[i] or b
			frame.sets[i] = frame.sets[i] or b
		end
	end

	if frame:HasScript("OnEnter") and frame.sets then
		local enter = frame:GetScript("OnEnter")
		if enter then
			frame:HookScript("OnEnter", onEnter)
		else
			frame:SetScript("OnEnter", onEnter)
		end
	end
	if frame:HasScript("OnLeave") and frame.sets then

		local leave = frame:GetScript("OnLeave")
		if leave then
			frame:HookScript("OnLeave", onLeave)
		else
			frame:SetScript("OnLeave", onLeave)
		end
	end
	
	hooksecurefunc(frame, "Layout", function(self)
		if (not self.sets) or InCombatLockdown() then
			return
		end
		local Strata = Addon.layers[ ceil(self.sets.frameStrata or 1)] or "MEDIUM"
		
		self:SetFrameStrata(Strata)
		self:SetFrameLevel(ceil(self.sets.frameLevel)+1)
		
		if self.drop then
			self.drop:SetFrameStrata(Strata)
			self.drop:SetFrameLevel(ceil(self.sets.frameLevel))
		end
		self:SetAlpha(self.sets.fadeAlpha/100 or 1)
		if self.key == "focus" then
			local s = true
			if self.sets.disableMouse and self.sets.disableMouse == true then
				s = false
			end
				
			
			self:EnableMouse(s)
		else
			self:EnableMouse(false)
		end

	end)
end

function Addon:Load()
	self.widgets = self.widgets or {}
	if #self.widgets == 0 then
		for i, info in pairs(Addon.Copy(pieces)) do
			local allow = true
			if info.func.GetRestrictions then
				local id, include = info.func.GetRestrictions()
				if (include == true) then
					if id ~= string.lower(self.id) then
						allow = false
					end
				elseif include == false then
					if id == string.lower(self.id) then
						allow = false
					end
				end
			end
			if allow then
				local widget, key = info.func:New(self), info.id
				if widget then
					widget.key = key
					if widget.defaults then
						self.sets[key] = self.sets[key] or {}
						self.check(widget.defaults, self.sets[key])
						widget.sets = self.sets[key]
						self[key] = widget
					end
					widget.ID = self.id
					tinsert(self.widgets, widget)
				end
			end
		end
	end
	for id, widget in pairs(self.widgets) do
		if widget.Load then
			widget:Load()
		end
	end
end

local globals = {
	advanced = {
		{
			name = "Frame Level",
			kind = "Slider",
			key = "frameLevel",
			min = 1,
			max = 100,
		},
		{
			name = "Frame Strata",
			kind = "Slider",
			key = "frameStrata",
			min = 1,
			max = 8,
		},
		{
			name = 'X Offset',
			kind = 'Slider',
			key = 'x',
			min = -400,
			max = 400,
		},
		{
			name = 'Y Offset',
			kind = 'Slider',
			key = 'y',
			min = -400,
			max = 400,
		},
		{
			name = 'Anchor',
			kind = 'Menu',
			key = 'anchor',
			table = {
				'TopLeft',
				'Top',
				'TopRight',
				'Right',
				'BottomRight',
				'Bottom',
				'BottomLeft',
				'Left',
				'Center',
			},
		},
	},
	basic = {
		{
			name = 'Scale',
			kind = 'Slider',
			key = 'scale',
			min = 50,
			max = 200,
			index = 1,
		},
		{
			name = 'Enable',
			kind = 'CheckButton',
			key = 'enable',
			index = 1,
		},
		{
			name = 'Opacity',
			kind = 'Slider',
			key = 'alpha',
			min = 0,
			max = 100,
		},
		{
			name = 'Faded Opacity',
			kind = 'Slider',
			key = 'fadeAlpha',
			min = 0,
			max = 100,
		},
	},
}

function AddGlobals(self)
	if self.ADDED then
		return
	end
	local options = Addon.Copy(self.Options)
	local def = self.defaults
	if def then
		def.frameLayer = 1
		def.frameStrata = 2
		self.sets.frameLayer = self.sets.frameLayer or def.frameLayer
		self.sets.frameStrata = self.sets.frameStrata or def.frameStrata
	end
	for title, menu in pairs(globals) do
		for i, option in pairs(menu) do
			option.panel = title
			if option.index then
				tinsert(options, option.index, option)
			else
				tinsert(options, option)
			end
		end
	end
	self.ADDED = true
	return options
end

local base = {
	CreateMenu = function(self, menu)
		if self.Options then
			local options = AddGlobals(self)
			Addon.Menu.NewMenu(self, menu, string.lower(self.title), Addon.Copy(options))
		end
	end,
	GetRestrictions = function() end
}

function Addon:NewWidget(name, frameType)
	local widget = Addon.master:CreateClass(frameType)
	widget.title = name
	widget.owner = self:GetParent()
 	tinsert(pieces, {id = string.lower(name), func = widget})
	Addon.Blend(Addon.Copy(base), widget)
 	return widget
end

function Addon:UpdateWidgets()
	if self.widgets then
		for id, widget in pairs(self.widgets) do
			if widget.Layout then
				Hook(widget)
				widget:Layout()
			end
		end
	end
end

local templates = {}
Addon.templates = templates

function Addon:NewTemplate(name)
	templates[name] = {}
	return templates[name]
end

function Addon:GetTemplate(template, name)
	local temp = templates[template]
	if temp then
		return Addon.Merge(CopyTable(templates[template]), Addon:NewWidget(string.lower(name), template))
	else
		print("Template does not exsist: "..template)
	end
end
